//
//  encryption.h
//  CryptTool
//
//  Created by Alexandre Colucci on 22/07/16.
//  Copyright © 2016 Alexandre Colucci. All rights reserved.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Include for mach-o structures
#include <mach-o/loader.h>

// Include for mach_task_self and vm_protect
#include <mach/mach.h>

// Includes for AES using CommonCrypto
#include <CommonCrypto/CommonCryptor.h>
#include <CommonCrypto/CommonKeyDerivation.h>



// AES key used
static char sAesKey[kCCKeySizeAES128+1] = { 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF };

// Buffer used to encrypt/decrypt the buffers. We don't encrypt/decrypt more than PAGE_SIZE bytes.
static uint8_t *sEncryptionBuffer[PAGE_SIZE + kCCBlockSizeAES128] = { 0 };


//
// Function that changes the virtual memory protections
//
// The __TEXT segment has its virtual memory protections set to not writable by default.
// In order to modify this code in memory, we need to change the virtual memory protections.
//
// This function sets the virtual memory protections to read, write and execute.
//
// Note that the function is forced inlined to ensure its code ends up in the correct section.
//
static inline void __attribute__((always_inline)) ChangeVirtualMemoryProtections(vm_address_t addr, vm_size_t size)
{
	kern_return_t returnValue = vm_protect(mach_task_self(), addr, size, false, VM_PROT_ALL);
	if ( returnValue != KERN_SUCCESS )
	{
		fprintf(stderr, "Error %d: Fail to change virtual memory protections\n", returnValue);
		exit(1);
	}
}

//
// Function that encrypts/decrypts len bytes from the source buffer to the destination buffer using AES.
//
// Note that the function is forced inlined to ensure its code ends up in the correct section.
//
static inline void __attribute__((always_inline)) EncryptDecryptBytes(const void *srcBuffer, void *dstBuffer, uint64_t len, CCOperation operation)
{
	// Encrypt/decrypt the data from the source buffer using the AES key
	size_t outLength = 0;
	CCCryptorStatus result = CCCrypt(operation,
                   kCCAlgorithmAES128,
                   kCCOptionPKCS7Padding,
                   sAesKey,
                   kCCKeySizeAES128,
                   NULL,
                   srcBuffer,
                   len,
                   sEncryptionBuffer,
                   len + kCCBlockSizeAES128,
                   &outLength);
	
	if (result == kCCSuccess)
	{
		// Copy the encrypted/decrypted data into the destination buffer
		memcpy(dstBuffer, sEncryptionBuffer, len);
	}
	else
	{
		fprintf(stderr, "Error %d: Could not %s the data\n", result, (operation == kCCEncrypt) ? "encrypt" : "decrypt");
		exit(1);
	}
}


//
// Function that encrypts/decrypts a buffer by chunks of PAGE_SIZE bytes.
// If there are less than PAGE_SIZE bytes, it does chunk of 16 bytes.
// This allows even simple program like an "Hello World" to be encrypted.
//
// Note that the function is forced inlined to ensure its code ends up in the correct section.
//
static inline void __attribute__((always_inline)) EncryptDecryptBuffer(const void *src, void *dst, uint64_t len, CCOperation operation)
{
	int32_t numberOfBytes = PAGE_SIZE;
	
	for (uint64_t i = 0; i < len; i += numberOfBytes)
	{
		if (numberOfBytes == PAGE_SIZE)
		{
			if (i + numberOfBytes > len && i + 16 <= len)
			{
				// There are less than PAGE_SIZE bytes but more than 16
				// we switch to 16 bytes
				numberOfBytes = 16;
			}
			else
			{
				// Less than 16 bytes
				// We stop the encryption/decryption
				break;
			}
		}
		else
		{
			if (i + numberOfBytes > len)
			{
				// Less than 16 bytes
				// We stop the encryption/decryption
				break;
			}
		}
		
		// Make the virtual memory permissions to writable for the destination buffer
		ChangeVirtualMemoryProtections((vm_address_t)(dst + i), numberOfBytes);
		
		// Encrypt/Decrypt the source buffer into the destination buffer
		EncryptDecryptBytes(src + i,  dst + i, numberOfBytes, operation);
	}
}

//
// Function that encrypts/decrypts a mach-o file
// It parses the mach-o file to find the (__TEXT, __text) section and encrypts/decrypts it.
//
// Note that the function is forced inlined to ensure its code ends up in the correct section.
//
static inline void __attribute__((always_inline)) EncryptDecryptMachoFile(struct mach_header_64 *machHeader, CCOperation operation)
{
	size_t segmentOffset = sizeof(struct mach_header_64);
	
	// For each load command of the mach-o file
	for (uint32_t i = 0; i < machHeader->ncmds; i++)
	{
		struct load_command *loadCommand = (struct load_command *)((uint8_t *) machHeader + segmentOffset);
		if(loadCommand->cmd == LC_SEGMENT_64)
		{
			// We found a 64-bit segment
			struct segment_command_64 *segCommand = (struct segment_command_64 *) loadCommand;
			
			// For each section in the 64-bit segment
			void *sectionPtr = (void *)(segCommand + 1);
			for (uint32_t nsect = 0; nsect < segCommand->nsects; ++nsect)
			{
				struct section_64 *section = (struct section_64 *)sectionPtr;
				fprintf(stderr, "Found the section (%s, %s)\n", section->segname, section->sectname);
				
				// Check if this is the __TEXT segment
				if (strncmp(segCommand->segname, SEG_TEXT, 16) == 0)
				{
					// Check if this is the __text section
					if (strncmp(section->sectname, SECT_TEXT, 16) == 0)
					{
						// This is the (__TEXT, __text) section.
						// We should encrypt/decrypt it in place.
						fprintf(stderr, "%s the (%s, %s) section\n", (operation == kCCEncrypt) ? "Encrypting" : "Decrypting", section->segname, section->sectname);
						EncryptDecryptBuffer((uint8_t *) machHeader + section->offset, (uint8_t *) machHeader + section->offset, section->size, operation);
						
					}
				}
				
				sectionPtr += sizeof(struct section_64);
			}
		}
		
		segmentOffset += loadCommand->cmdsize;
	}
}
